
//=====================================================================
// WifiManager AP IP-address: 192.168.4.1
//=====================================================================


//=====================================================================
// Libraries, constants and globals
//=====================================================================
#include <FS.h>
#include <ESP8266WiFi.h>
#include <ESP8266WiFiMulti.h>
#include <ESP8266WebServer.h>
#include <WebSocketsServer.h>
#include <WiFiManager.h>

ESP8266WiFiMulti wifiMulti;        // Create an instance of the ESP8266WiFiMulti class, called 'wifiMulti'
ESP8266WebServer server(80);       // Create a webserver object that listens for HTTP request on port 80
WebSocketsServer webSocket(81);    // create a websocket server on port 81
WiFiManager wifiManager;

File fsUploadFile;                 // a File variable to temporarily store the received file

const bool DEBUG        = true;   // "true":print to serial  "false":debugging switched off

const int WIFILED       = LED_BUILTIN; //D4;   // internal LED; WIFI flag
const int GPIO_MOT_M1A  = D1;   // GPIO for "motor A driver address 1"
const int GPIO_MOT_M2A  = D2;   // GPIO for "motor A driver address 2"
const int GPIO_FRONTL   = D5;   // GPIO for "front lights"
const int GPIO_BACKLI   = D6;   // GPIO for "back lights"
const int GPIO_FUNCTI   = D7;   // GPIO for "additional function"

const char* ID_ADDRS = "A";   // identifier for websocket address
const char* ID_INFOS = "I";   // identifier for get all informations
const char* ID_STOPP = "E";   // identifier for signal "EMERGENCY STOP"
const char* ID_SPEED = "S"; 	// identifier for signal "SPEED"
const char* ID_DIREC = "D"; 	// identifier for signal "DIRECTION"
const char* ID_LIGHT = "L"; 	// identifier for signal "LIGHTS"
const char* ID_FUNCT = "F";   // identifier for signal "FUNCTION"
const char* ID_WIFIS = "W";   // identifier for signal "WIFI SIGNAL STRENGTH"
const char* ID_LNAME = "N";   // identifier for signal "LOCOMOTIVE NAME"
const char* ID_SWMOT = "M";   // identifier for signal "SWITCH MOTOR DIRECTION CONFIG"
const char* ID_SWLIG = "G";   // identifier for signal "SWITCH LIGHTS DIRECTION CONFIG"
const char* ID_RESET = "R";   // RESET WIFI SETTINGS
const char* ID_PICTU = "P";   // receiving PICTURE data:image/jpeg;base64
const char* ID_PICTX = "X";   // send PICTURE data:image/jpeg;base64

String c_loconame;            // name of locomotive
String c_icon;                // icon path
bool   c_motor;               // config motor direction (TRUE = invert)
bool   c_lights;              // config light direction (TRUE = invert)


int ws_num        = 0;    // connected websocket clients

int		SPEED = 0;			  	// engine speed
bool	FWD   = false;		  // direction forward/backward
bool	LIGHT = false;			// lights on/off
bool	FRONT = false;			// front lights on/off
bool	BACKL = false;			// back lights on/off
bool	FUNCT = false;			// function on/off

// void handleFileUpload();                // upload a new file to the SPIFFS

//=====================================================================
// SETUP
//=====================================================================
void setup() {
  Serial.begin(115200);		// Start the Serial communication to send messages to the computer
  delay(10);
  showDebug("\r\n");
  showDebug("----------------------------");

  showDebug("=> start SPIFFS ------------");
  startSPIFFS();      // Start the SPIFFS and list all contents

  showDebug("=> load configuration ------");
  initPARAMS();      // load configuration and init parameters

  showDebug("=> start Wifi Manager ------");
  initWifiManager();  // Start WifiManager

  showDebug("=> initilize GPIOs ---------");
  initGPIOs();        // initilizing the GPIO outputs

  showDebug("=> start WebSocket ---------");
  startWebSocket();			// Start a WebSocket server

  showDebug("=> start Server ------------");
  startServer();			// Start a HTTP server with a file read handler and an upload handler

  showDebug("----------------------------");
}


//=====================================================================
// LOOP
//=====================================================================
void loop() {
  webSocket.loop();			    // constantly check for websocket events
  server.handleClient();	  // run the server
}


//=====================================================================
// Setup functions
//=====================================================================

void configModeCallback (WiFiManager *myWiFiManager) {
  showDebug("Entered config mode");
  Serial.println( WiFi.softAPIP() );
  //if you used auto generated SSID, print it
  showDebug(myWiFiManager->getConfigPortalSSID());
}

void initWifiManager() {
  //WiFiManager
  //Local intialization. Once its business is done, there is no need to keep it around
//  WiFiManager wifiManager;
  //reset settings - for testing
  //wifiManager.resetSettings();
  //set callback that gets called when connecting to previous WiFi fails, and enters Access Point mode
  wifiManager.setAPCallback(configModeCallback);  
  //fetches ssid and pass and tries to connect
  //set timeout for autoConnect to latest SSID configuration.
  //if it doesnt connect automatically, then it will restart the ESP and will act as AccessPoint.
  //         wifiManager.setConnectTimeout(180);
  //         wifiManager.setTimeout(180);
  //if it does not connect it starts an access point with the specified name
  //here  "AutoConnectAP"
  //and goes into a blocking loop awaiting configuration
  wifi_station_set_hostname((char *)c_loconame.c_str());
  if(!wifiManager.autoConnect()) {
    showDebug("failed to connect and hit timeout");
    //reset and try again, or maybe put it to deep sleep
    ESP.reset();
    delay(1000);
    //resetWifi();
  }  
  showDebug("WifiManager started..."); 
  digitalWrite(WIFILED, LOW);  // turn WIFI LED "ON" -> WIFI connected
}


void initPARAMS() {
  loadConfig();
  showDebug(String("\t c_loconame: ") + c_loconame);
  showDebug(String("\t c_icon: ") + c_icon);
  showDebug(String("\t c_motor: ") + bool2str(c_motor));
  showDebug(String("\t c_lights: ") + bool2str(c_lights));
  delay(500);
  WiFi.hostname(c_loconame);
}




void startSPIFFS() { // Start the SPIFFS and list all contents
  SPIFFS.begin();                             // Start the SPI Flash File System (SPIFFS)
  showDebug("SPIFFS started. Contents:");
  if (DEBUG) {
    Dir dir = SPIFFS.openDir("/");
    while (dir.next()) {                      // List the file system contents
      String fileName = dir.fileName();
      size_t fileSize = dir.fileSize();
      Serial.printf("\tFS File: %s, size: %s\r\n", fileName.c_str(), formatBytes(fileSize).c_str());
    }
    showDebug("\n");
  }
}



void startWebSocket() { // Start a WebSocket server
  webSocket.begin();                          // start the websocket server
  webSocket.onEvent(webSocketEvent);          // if there's an incomming websocket message, go to function 'webSocketEvent'
  Serial.println("WebSocket server started.");
}



void startServer() { // Start a HTTP server with a file read handler and an upload handler
  server.on("index.html",  HTTP_POST, []() {  // If a POST request is sent to the /edit.html address,
    server.send(200, "text/plain", "");
  }, handleFileUpload);                       // go to 'handleFileUpload'
  server.on("/upload", HTTP_POST,             // if the client posts to the upload page
    [](){ server.send(200); },                // Send status 200 (OK) to tell the client we are ready to receive
    handleFileUpload                          // Receive and save the file
  );

  server.on("/genericArgs", handleGenericArgs);
  server.on("/specificArgs", handleSpecificArg);

  server.onNotFound(handleNotFound);          // if someone requests any other file or page, go to function 'handleNotFound'
  // and check if the file exists
  server.begin();                             // start the HTTP server
  Serial.println("HTTP server started.");
}



void initGPIOs() {	// initilizing the GPIO outputs
  pinMode(WIFILED,      OUTPUT);  digitalWrite(WIFILED, HIGH);                // turn WIFI LED "OFF"
  pinMode(GPIO_MOT_M1A, OUTPUT);  digitalWrite(GPIO_MOT_M1A, 0);              // init motor A driver address 1
  pinMode(GPIO_MOT_M2A, OUTPUT);  digitalWrite(GPIO_MOT_M2A, 0);              // init motor A driver address 2
  pinMode(GPIO_FRONTL,  OUTPUT);  digitalWrite(GPIO_FRONTL, bool2int(FRONT)); // turn off the front lights
  pinMode(GPIO_BACKLI,  OUTPUT);  digitalWrite(GPIO_BACKLI, bool2int(BACKL)); // turn off the back lights
  pinMode(GPIO_FUNCTI,  OUTPUT);  digitalWrite(GPIO_FUNCTI, bool2int(FUNCT)); // turn off the function
  Serial.println("GPIO's initilized.");
}


//=====================================================================
// Server handlers
//=====================================================================

void handleNotFound() { // if the requested file or page doesn't exist, return a 404 not found error
  if (!handleFileRead(server.uri())) {        // check if the file exists in the flash memory (SPIFFS), if so, send it
    server.send(404, "text/plain", "404: File Not Found");
  }
}

void handleGenericArgs() { //Handler
  String message = "Number of args received:";
  message += server.args();            //Get number of parameters
  message += "\n";                     //Add a new line
  for (int i = 0; i < server.args(); i++) {
    //Include the current iteration value
    message += "Arg nº" + (String)i + " –> ";
    //Get the name of the parameter
    message += server.argName(i) + ": ";
    //Get the value of the parameter
    message += server.arg(i) + "\n";
  }
  //Response to the HTTP request
  server.send(200, "text/plain", message);
}

void handleSpecificArg() {
  String message = "";
  if (server.arg("Temperature") == "") {   //Parameter not found
    message = "Temperature Argument not found";
  } else {    //Parameter found
    message = "Temperature Argument = ";
    //Gets the value of the query parameter
    message += server.arg("Temperature");
  }
  //Returns the HTTP response
  server.send(200, "text/plain", message);
}


bool handleFileRead(String path) { // send the right file to the client (if it exists)
  showDebug("handleFileRead: " + path);
  if (path.endsWith("/")) path += "index.html";          // If a folder is requested, send the index file
  String contentType = getContentType(path);             // Get the MIME type
  String pathWithGz = path + ".gz";
  if (SPIFFS.exists(pathWithGz) || SPIFFS.exists(path)) { // If the file exists, either as a compressed archive, or normal
    if (SPIFFS.exists(pathWithGz))                         // If there's a compressed version available
      path += ".gz";                                         // Use the compressed verion
    File file = SPIFFS.open(path, "r");                    // Open the file
    size_t sent = server.streamFile(file, contentType);    // Send it to the client
    file.close();                                          // Close the file again
    showDebug(String("\tSent file: ") + path);
    return true;
  }
  Serial.println(String("\tFile Not Found: ") + path);   // If the file doesn't exist, return false
  return false;
}


void handleFileUpload(){ // upload a new file to the SPIFFS
  HTTPUpload& upload = server.upload();
  if(upload.status == UPLOAD_FILE_START){
    String filename = upload.filename;
    if(!filename.startsWith("/")) filename = "/"+filename;
    Serial.print("handleFileUpload Name: "); Serial.println(filename);
    fsUploadFile = SPIFFS.open(filename, "w");            // Open the file for writing in SPIFFS (create if it doesn't exist)
    filename = String();
  } else if(upload.status == UPLOAD_FILE_WRITE){
    if(fsUploadFile)
      fsUploadFile.write(upload.buf, upload.currentSize); // Write the received bytes to the file
  } else if(upload.status == UPLOAD_FILE_END){
    if(fsUploadFile) {                                    // If the file was successfully created
      fsUploadFile.close();                               // Close the file again
      Serial.print("handleFileUpload Size: "); Serial.println(upload.totalSize);
      server.sendHeader("Location","/success.html");      // Redirect the client to the success page
      server.send(303);
    } else {
      server.send(500, "text/plain", "500: couldn't create file");
    }
  }
}


String bool2str(bool state) {
  String strval = state == true ? "1" : "0";
  return strval;
}

String bool2str(String ID, bool val) {
  // String strval = val == true ? "1" : "0";
  return (String)ID + bool2str(val);
}

int bool2int(bool state) {
  if (state == true) {
    return 1;
  } else {
    return 0;
  }
}



void webSocketEvent(uint8_t num, WStype_t type, uint8_t * payload, size_t length) { // When a WebSocket message is received
  switch (type) {
    case WStype_DISCONNECTED:             // if the websocket is disconnected
      showDebug((String) printf("[%u] Disconnected!\n", num));
      ws_num = webSocket.connectedClients();
      break;
    case WStype_CONNECTED: {              // if a new websocket connection is established
        IPAddress ip = webSocket.remoteIP(num);
        Serial.printf("[%u] Connected from %d.%d.%d.%d url: %s\n", num, ip[0], ip[1], ip[2], ip[3], payload);
        String connReply = (String)ID_INFOS + ID_LNAME + c_loconame +"|"+ getWifiSignal(ID_WIFIS) + "|" + ID_ADDRS + (String)num + "|" + (String)ID_SPEED + (String)SPEED + "|" + bool2str(ID_DIREC, FWD) + "|" 
            + bool2str(ID_LIGHT, LIGHT) + "|" +  bool2str(ID_FUNCT, FUNCT) + "|" +  bool2str(ID_SWMOT, c_motor) + "|" +  bool2str(ID_SWLIG, c_lights);
        showDebug("connection reply: " + connReply);
        webSocket.sendTXT(num, connReply);
        ws_num = webSocket.connectedClients();
      }
      break;
    case WStype_TEXT:                     // if new text data is received
      //showDebug( (String) printf("[%u] get Text: %s\n", num, payload));
      char * PL = (char *) &payload[0];
      char * PV = (char *) &payload[1];
      //showDebug((String) sprintf("pay-value: %s\n", PV));
      String PLstr = &PL[0];
      String key = (String)PLstr.charAt(0);
      showDebug("key= " + key + "\n");
      showDebug("PLstr= " + PLstr + "\n");

      //------------------------------------------------------------------ ask for information
      if (key == ID_INFOS) {
        String connReply = (String)ID_INFOS + ID_LNAME + c_loconame +"|"+ getWifiSignal(ID_WIFIS) + "|" + ID_ADDRS + (String)num + "|" + (String)ID_SPEED + (String)SPEED + "|" 
           + bool2str(ID_DIREC, FWD) + "|" + bool2str(ID_LIGHT, LIGHT) + "|" +  bool2str(ID_FUNCT, FUNCT) + "|" +  bool2str(ID_SWMOT, c_motor) + "|" +  bool2str(ID_SWLIG, c_lights);
        showDebug("connection reply: " + connReply);
        webSocket.sendTXT(num, connReply);

        //--------------------------------------------------------------- change the engine speed
      } else if (key == ID_SPEED) {
        SPEED = (int) strtol((const char *) &payload[1], NULL, 10);
        ctrl_Motor();
        String strval = ID_SPEED + (String)SPEED;
        showDebug("broadcast=" + strval + "\n");
        webSocket.broadcastTXT(strval);
        //---------------------------------------------------------------------- "WIFI SIGNAL STRENGTH"
      } else if (key == ID_WIFIS) {
        String strval = getWifiSignal(ID_WIFIS);
        showDebug("WIFI SIGNAL = " + strval + " dB\n");
        webSocket.broadcastTXT(strval);
        //---------------------------------------------------------------------- "LOCOMOTIVE NAME"
      } else if (key == ID_LNAME) {
        String Lname = (const char *) &payload[1];
        c_loconame = Lname;
        // saveConfig...
        saveConfig();
        // broadcast new name to all clients
        String strval = ID_LNAME + Lname;
        webSocket.broadcastTXT(strval);
        //---------------------------------------------------------------------- "EMERGENCY STOP"
      } else if (key == ID_STOPP) {
        SPEED = 0;
        ctrl_Motor();
        String strval = (String)ID_SPEED + "0";
        showDebug("broadcast=" + strval + "\n");
        webSocket.broadcastTXT(strval);
        //------------------------------------------------------------------ change the direction
      } else if (key == ID_DIREC) {
        if (SPEED < 1) {   // only toggle direction if SPEED < 1
          String val = (String)PLstr.charAt(1);
          FWD = (val == "1");
          ctrl_Lights();
          String strB = FWD == true ? "1" : "0";
          String strval = ID_DIREC + strB;
          webSocket.broadcastTXT(strval);
        }
        //--------------------------------------------------------------- change the light status
      } else if (key == ID_LIGHT) {
        LIGHT  = !LIGHT;
        ctrl_Lights();
        String strB = LIGHT == true ? "1" : "0";
        String strval = ID_LIGHT + strB;
        webSocket.broadcastTXT(strval);
        //---------------------------------------------------------- change the functional status
      } else if (key == ID_FUNCT) {
        FUNCT = !FUNCT;
        ctrl_Function();
        String strB = FUNCT == true ? "1" : "0";
        String strval = ID_FUNCT + strB;
        webSocket.broadcastTXT(strval);
        break;
        //------------------------------------------------------------------- reset WIFI settings
      } else if (key == ID_RESET) {
          // stop engine before loosing connection
          SPEED = 0;
          ctrl_Motor();
          // Reset WifiManager setting and reboot ESP8266
          resetWifi();
        break;
        //----------------------------------------------------------------- receive PICTURE data
      } else if (key == ID_PICTU) {
        showDebug("==========================");
        c_icon = (String) PV;
        showDebug("received image data=" + c_icon + "\n");
        saveConfig();
        showDebug("c_icon = " + c_icon + "\n");
        // broadcast new icon...
        String strval = ID_PICTX + c_icon;
        showDebug("broadcast icon = " + strval + "\n");
        webSocket.broadcastTXT(strval);
        showDebug("==========================");
        //----------------------------------------------------------------- request PICTURE data
      } else if (key == ID_PICTX) {
        String strval = ID_PICTX + c_icon;
        webSocket.broadcastTXT(strval);   
        //----------------------------------------------------------------- request CONFIG MOTOR DIRECTION INVERT
      } else if (key == ID_SWMOT) {
        String cmot = (const char *) &payload[1];
        c_motor = (cmot == "1"); 
        saveConfig();
        String strval = ID_SWMOT + cmot;
        webSocket.broadcastTXT(strval);
        break;              
        //----------------------------------------------------------------- request CONFIG LIGHT DIRECTION INVERT
      } else if (key == ID_SWLIG) {
        String clight = (const char *) &payload[1];
        c_lights = (clight == "1"); 
        saveConfig();
        String strval = ID_SWLIG + clight;
        webSocket.broadcastTXT(strval);
        break;        
      }
  }
}



//=====================================================================
// Helper functions
//=====================================================================



// get WIFI signal strength in Dezibel
String getWifiSignal(String ID) {
  long rssi = abs(WiFi.RSSI());
  char strval[40];
  sprintf(strval,"%ld",rssi);
  return (String)ID + strval;
}

void resetWifi() {
  WiFi.disconnect();
  wifiManager.resetSettings();
  ESP.eraseConfig();
  ESP.reset();
  ESP.restart();
  yield();
}


// list all "l_*.png files
String getLocoImgList() {
  String imgList = "";
  String fname   = "";
  Dir dir = SPIFFS.openDir("");
  while (dir.next()) {
    fname = dir.fileName();
    if (fname.endsWith(".png")) {
      if (fname.startsWith("/l_")) {
        Serial.print("getIconList: ");
        Serial.println(dir.fileName());
        imgList = imgList + ";" + dir.fileName();
      }
    }
  }
  if (imgList.length() > 1) {
    imgList[0] = '|';
  }
  Serial.print("getIconList (done): ");
  Serial.println(imgList);
  return imgList;
}


String split(String data, char separator, int index) {
  int found = 0;
  int strIndex[] = { 0, -1 };
  int maxIndex = data.length();

  for (int i = 0; i <= maxIndex && found <= index; i++) {
    if (data.charAt(i) == separator || i == maxIndex) {
      found++;
      strIndex[0] = strIndex[1] + 1;
      strIndex[1] = (i == maxIndex) ? i + 1 : i;
    }
  }
  return found > index ? data.substring(strIndex[0], strIndex[1]) : "";
}  // END


void saveConfig() {  // save the configuration to SPIFFS
  String config = c_loconame +"|"+ c_icon +"|"+ bool2str(c_motor) +"|"+ bool2str(c_lights) +"|";
  File f = SPIFFS.open("/config.dat", "w");
  if (!f) {
    showDebug("failed to open configuration file!!!");
    return;
  } else {
    int fw = f.print(config);
    if (fw < 1) {
      showDebug("configuration not saved!");
    } else {
      showDebug("configuration saved:" + config);  
    }
    f.close();
  }
}

void loadConfig() {   // load configuration from SPIFFS
  File f = SPIFFS.open("/config.dat", "r");
  if (!f) {
    Serial.println("failed to read configuration!!!");
    c_loconame = "dummy";
    c_icon     = "";
    c_motor    = false;   // config motor direction (TRUE = invert)
    c_lights   = false;   // config light direction (TRUE = invert)
  } else {
    String line = f.readStringUntil('\n');
    f.close();
    showDebug("configuration successfully read:" + line);
    c_loconame = split(line, '|', 0);          // name of locomotive
    c_icon     = split(line, '|', 1);          // icon path
    c_motor    = (split(line, '|', 2) == "1"); // config motor direction (TRUE = invert)
    c_lights   = (split(line, '|', 3) == "1"); // config light direction (TRUE = invert)
  }
}

void showDebug(String msg) {
  if (DEBUG == true) {
    Serial.println(msg);
  }
}

String formatBytes(size_t bytes) { // convert sizes in bytes to KB and MB
  if (bytes < 1024) {
    return String(bytes) + "B";
  } else if (bytes < (1024 * 1024)) {
    return String(bytes / 1024.0) + "KB";
  } else if (bytes < (1024 * 1024 * 1024)) {
    return String(bytes / 1024.0 / 1024.0) + "MB";
  }
}


void ctrl_Motor() {
  // check limits ---------------------
  if (SPEED <    0) { SPEED =    0; }
  if (SPEED > 1023) { SPEED = 1023; }
  // control speed --------------------
  if (SPEED < 100) {  //----------- STOPP
      digitalWrite(GPIO_MOT_M1A, LOW);
      digitalWrite(GPIO_MOT_M2A, LOW);
  } else {  //-------------------- SPIN
    // logic check if config inverts
    bool direc = FWD;
    if (c_motor) { direc = ! FWD; }
    // now speed up the motor in the correct direction
    if (direc) {
      digitalWrite(GPIO_MOT_M1A, LOW);
      analogWrite( GPIO_MOT_M2A, SPEED);
    } else {
      analogWrite( GPIO_MOT_M1A, SPEED);
      digitalWrite(GPIO_MOT_M2A, LOW);
    }
  }
}


void ctrl_Lights() {
  if (LIGHT) {     //---------------------------------- lights are ON
    // logic check if config inverts
    bool direc = FWD;
    if (c_lights) { direc = ! FWD; }
    // now let the light shine
    if (direc) {
      digitalWrite(GPIO_FRONTL, 1);	// turn ON the front lights
      digitalWrite(GPIO_BACKLI, 0);	// turn OFF the back lights
    } else {
      digitalWrite(GPIO_FRONTL, 0);	// turn OFF the front lights
      digitalWrite(GPIO_BACKLI, 1);	// turn ON the back lights
    }
  } else {     //------------------------------------- lights are OFF
    digitalWrite(GPIO_FRONTL, 0);	// turn OFF the front lights
    digitalWrite(GPIO_BACKLI, 0);	// turn OFF the back lights
  }
}


void ctrl_Function() {
  if (FUNCT) {
    digitalWrite(GPIO_FUNCTI, HIGH); // turn ON
  } else {
    digitalWrite(GPIO_FUNCTI, LOW);  // turn OFF
  }
}

String getContentType(String filename) { // determine the filetype of a given filename, based on the extension
  String dataType = "text/plain";
  if (filename.endsWith("/")) filename += "index.htm";
  if (filename.endsWith(".src")) filename = filename.substring(0, filename.lastIndexOf("."));
  else if (filename.endsWith(".html")) dataType = "text/html";
  else if (filename.endsWith(".htm")) dataType = "text/html";
  else if (filename.endsWith(".css")) dataType = "text/css";
  else if (filename.endsWith(".js")) dataType = "application/javascript";
  else if (filename.endsWith(".png")) dataType = "image/png";
  else if (filename.endsWith(".gif")) dataType = "image/gif";
  else if (filename.endsWith(".jpg")) dataType = "image/jpeg";
  else if (filename.endsWith(".ico")) dataType = "image/x-icon";
  else if (filename.endsWith(".xml")) dataType = "text/xml";
  else if (filename.endsWith(".pdf")) dataType = "application/pdf";
  else if (filename.endsWith(".zip")) dataType = "application/zip";
  return dataType;
}