var connection = new ReconnectingWebSocket('ws://' + location.hostname + ':81/', ['arduino']);
connection.reconnectInterval = 1000;  // check every second
connection.maxReconnectAttempts = 60; // try a minute until giving up
// var connection = new WebSocket('ws://' + location.hostname + ':81/', ['arduino']);


var wsEnable       = true; // SPEED SLIDER ENABLE
var iconEvListener = true;


function wsSwitchButton(ID, val) {
	btn = document.getElementById(ID);
	if (btn) { btn.setAttribute("data", val); }
}

function iconWIFI(id, disp) {
	icon = document.getElementById(id);
	if (icon) { icon.style.display = disp; }
}

function clickResetWifi() {
	var conf = confirm("Do you really want to reset your WIFI settings?\nThis will take a few seconds.");
	if (conf == true) {
		sendCtrl("R", "");
		alert("The IP address in the ESP-network will be 192.168.4.1");
	}
}


function updateWIFIsignal(val) {
	strength = Number(val);
	if (strength > 65) {
		iconWIFI("W3", "block");
		iconWIFI("W2", "none");
		iconWIFI("W1", "none");
		iconWIFI("W0", "none");
	} else if (strength > 50) {
		iconWIFI("W3", "none");
		iconWIFI("W2", "block");
		iconWIFI("W1", "none");
		iconWIFI("W0", "none");
	} else if (strength > 25) {
		iconWIFI("W3", "none");
		iconWIFI("W2", "none");
		iconWIFI("W1", "block");
		iconWIFI("W0", "none");
	} else {
		iconWIFI("W3", "none");
		iconWIFI("W2", "none");
		iconWIFI("W1", "none");
		iconWIFI("W0", "block");
	}
}

function updateLocoName(val) {
	labelName = document.getElementById("lbl_loconame");
	labelName.innerHTML = val;
	document.title = val;
}

function wsConfigCB(checkBox, value) {
	if (checkBox == "M") { document.getElementById("swMotor").checked  = (value == "1"); }
	if (checkBox == "G") { document.getElementById("swLights").checked = (value == "1"); }
}

function handleRecvKey(key, value) {
	switch(key) {
		case "L":   // mark button for "LIGHT"
			wsSwitchButton('L', value);
			break;
		case "F":   // mark button for "FUNCTION"
			wsSwitchButton('F', value);
			break;
		case "W":   // modify icon for "WIFI signal strength"
			updateWIFIsignal(value);
			break;
		case "D":   // mark buttons for "FORWARD"
			if (value == "0") {
				wsSwitchButton("V", "1");
				wsSwitchButton("Z", "0");
			} else{
				wsSwitchButton("V", "0");
				wsSwitchButton("Z", "1");
			}
			break;
		case "S":    // move the slider remotely to SPEED
			wsEnable = false;
			$slider.val(value).change();
			wsEnable = true;
			break;
		case "N":   // update locomotive name
			updateLocoName(value);
			break;
		case "M":   // configuration switch motor direction
			wsConfigCB(key, value);
			break;
		case "G":   // configuration switch lights direction
			wsConfigCB(key, value);
			break;
	}	
}

/**********************************************
  ============================
    WebSocket Command Set
  ============================

	L	LIGHTS				[0;1]
	F	FUNCTION			[0;1]
	D	DIRECTION			[1=FORWARD; 0=BACKWARDS]
	V	FORWARD				[ - ]	// outdated
	Z	BACKWARDS			[ - ]	// outdated
	E	EMERGENCY STOPP		[ - ]
	I	INFO				[ - ]
	S	SPEED				[0..1023]
	N	LOCOMOTIVE NAME		[ASCII]
	M	CONFIG MOTOR DIR	[0;1]
	G	CONFIG LIGHT DIR	[0;1]
	R	RESET WIFI SETTINGS	[ - ]
	P	send PICTURE data:image/jpeg;base64
	X	receive PICTURE data:image/jpeg;base64

  ----------------------------
  EXAMPLE "I" received information
  IW-52|L0|F0|D1|S998|NStainz|M0|C1
	WIFI SIGNAL = -52dB
	LIGHTS      = "off"
	FUNCTION    = "off"
	DIRECTION   = "FORWARD"
	SPEED       = 998
	NAME        = "Stainz"
	CONFIGMOT   = "off"
	CONFIGLIGHT = "on"
  
**********************************************/
function handleWebsocketMsg(msg) {
	key = msg.charAt(0);
	value = msg.substr(1);
	switch(key) {
		case "I":    // receive initial information
			wsEnable = false;
			valArr = value.split('|');
			valArr.forEach(function(item) {
				k = item.charAt(0);
				v = item.substr(1);
				handleRecvKey(k, v);
			});
			wsEnable = true;
			break;
		case "X":    // receive picture data
			document.getElementById("icon").src = value;
			break;
		default:   // handle key directly
			handleRecvKey(key, value);
			break;
	}
}

function warnIcon(OnOff) {
	var icon = document.getElementById("warning");
	if (OnOff) {
		icon.style.display  = "block";
	} else {
		icon.style.display  = "none";
	}
}

connection.onopen = function () {
  connection.send('Connect ' + new Date());
  warnIcon(false);
};

connection.onerror = function (error) {
  console.log('WebSocket Error ', error);
  warnIcon(true);
};

connection.onmessage = function (e) {
  console.log('Server: ', e.data);
  handleWebsocketMsg(e.data);
};

connection.onclose = function () {
  console.log('WebSocket connection closed');
  warnIcon(true);
};

function btnClick(elem) {
	var id = elem.id;
	var value = elem.getAttribute("data");
	switch(id) {
		case "V":    // FORWARD button pressed
			sendCtrl("D", "0");
			break;
		case "Z":    // BACKWARDS button pressed
			sendCtrl("D", "1");
			break;
		default:   // handle key directly
			sendCtrl(id, value);
			break;
	}
}

function nameChange(elem) {
	/*
	var id = elem.id;
	var value = elem.value;
	sendCtrl("N", value);
	*/
}


function getDataUrl(img) {
   // Create canvas
   const canvas = document.createElement('canvas');
   const ctx = canvas.getContext('2d');
   // Set width and height
   canvas.width = img.width;
   canvas.height = img.height;
   // Draw the image
   ctx.drawImage(img, 0, 0);
   return canvas.toDataURL('image/jpeg');
}

function iconChanged() {
	var image = document.getElementById("upload-Image");
	image.src=event.target.result;
	// call webserver
	  intervalId = window.setInterval(function() {
		document.getElementById("icon").src = canvas.toDataURL();
		const dataUrl = getDataUrl(img);
		sendCtrl("P",dataUrl);
		window.clearInterval(intervalId);
		alert("setInterval icon changed");
	  }, 500);

	//var img = document.getElementById("icon");
	//const dataUrl = getDataUrl(img);
	//sendCtrl("P",dataUrl);
}

$( document ).ready(function() {
	var slider = document.getElementById("slider");
	if (slider) { slider.style.display = "block"; }
	connection.send("I");
	
	$('.w3-input w3-border').keypress(function (e) {  // Accept only alpha numerics, no special characters 
        var regex = new RegExp("^[a-zA-Z0-9 ]+$");
        var str = String.fromCharCode(!e.charCode ? e.which : e.charCode);
        if (regex.test(str)) {
            return true;
        }

        e.preventDefault();
        return false;
    });
});

// send Loco Name while loosing the focus of config input
$( "p" )
	.focusout(function() {
		var edit = document.getElementById("edit_loconame");
		if (edit) { 
			sendCtrl("N", edit.value); 
		}
	}
);

function sendCtrl(id, value) {
	if (wsEnable) {
		var cmd = id + value;  // build information out of "ID" and "VALUE"
		connection.send(cmd);  // send information to ESP8266
		console.log('CTRL send: "' +cmd+ '"');
	}
}

setInterval(
	function(){
		connection.send("W");
	}, 5000);

